;; Generated package description from which-key.el  -*- no-byte-compile: t -*-
(define-package "which-key" "3.3.2.0.20200904.200731" "Display available keybindings in popup" '((emacs "24.4")) :authors '(("Justin Burkett" . "justin@burkett.cc")) :maintainer '("Justin Burkett" . "justin@burkett.cc") :url "https://github.com/justbur/emacs-which-key")
